"""Logger is in charge of definning the output format based on CI expected output."""
from ostorlab.cli import console as cli_console
from ostorlab.cli.ci_scan.run.ci_logger import logger

console = cli_console.Console()


class Logger(logger.Logger):
    """Logger is in charge of printing the results based on the CI expected output."""

    def info(self, message: str) -> None:
        """Print Info messages.

        Args:
            message: message to print.
        """
        console.info(message)

    def error(self, message: str) -> None:
        """Print Error messages.

        Args:
            message: message to print.
        """
        console.error(message)


    def output(self, name: str, value: str) -> None:
        """Pass an output to the next step of the CI.

        Args:
            name: name of the output to pass to the next step.
            value: value of the output.
        """
        console.success(f'Output: {name}:{value}')



"""Runtime agent module gives access to methods to create the specific agent configurations & run the agent.

Usage
    agent_runtime = AgentRuntime(agent_settings, runtime_name, docker_client, mq_service, redis_service)
    agent_service = agent_runtime.create_agent_service(network_name, extra_configs)
"""
import io
import logging
import hashlib
from typing import List, Optional

import docker
from docker import constants
from docker import errors
from docker.types import services as docker_types_services

from ostorlab import configuration_manager
from ostorlab import exceptions
from ostorlab.agent import definitions as agent_definitions
from ostorlab.runtimes import definitions
from ostorlab.runtimes.local.services import mq
from ostorlab.runtimes.local.services import redis

logger = logging.getLogger(__name__)

MOUNT_VARIABLES = {
    '$CONFIG_HOME': str(configuration_manager.OSTORLAB_PRIVATE_DIR)
}

HEALTHCHECK_HOST = '0.0.0.0'
HEALTHCHECK_PORT = 5000
SECOND = 1000000000
HEALTHCHECK_RETRIES = 5
HEALTHCHECK_TIMEOUT = 10 * SECOND
HEALTHCHECK_START_PERIOD = 2 * SECOND
HEALTHCHECK_INTERVAL = 2 * SECOND


class Error(exceptions.OstorlabError):
    """Base Error."""


class MissingAgentDefinitionLabel(Error):
    """Agent definition label is missing from the agent image. This is likely due to the agent built directly with a
     docker command and not agent build."""


def _parse_mount_string_windows(string):
    """Handles parsing mounts on Windows OS."""
    parts = string.split(':')
    if len(parts) == 1:
        # This supposes only having a target like /root.
        return docker_types_services.Mount(target=parts[0], source=None)
    elif len(parts) == 2:
        # This supposes having a source linux style and target. Like /var/run/docker.sock:/var/run/docker.sock.
        target = parts[1]
        source = parts[0]
        mount_type = 'volume'
        if source.startswith('/'):
            # Paths likes /var/run/docker.sock map to //var/run/docker.sock on windows.
            source = f'/{source}'
            mount_type = 'bind'
        return docker_types_services.Mount(target, source, read_only=False, type=mount_type)
    elif len(parts) == 3:
        # This supposes two cases. First case is windows style with linux, like C:/Users/bob:/root. The second
        # case is /root:/root:ro
        if parts[2] in ('ro', 'rw', 'z', 'Z'):
            target = parts[1]
            source = parts[0]
            mount_type = 'volume'
            if source.startswith('/'):
                # Paths likes /var/run/docker.sock map to //var/run/docker.sock on windows.
                source = f'/{source}'
                mount_type = 'bind'
            read_only = not parts[2] == 'rw'
            return docker_types_services.Mount(target, source, read_only=read_only, type=mount_type)
        else:
            target = parts[2]
            source = ':'.join(parts[:2])
            source = source.replace('\\', '/')
            mount_type = 'bind'
            return docker_types_services.Mount(target, source, read_only=False, type=mount_type)
    elif len(parts) == 4:
        # This covers the case C:/Users/bob:/root:ro
        target = parts[2]
        source = ':'.join(parts[:2])
        source = source.replace('\\', '/')
        mount_type = 'bind'
        read_only = not (len(parts) == 3 or parts[3] == 'rw')
        return docker_types_services.Mount(target, source, read_only=read_only, type=mount_type)
    else:
        raise errors.InvalidArgument(
            f'Invalid mount format "{string}"'
        )


def _parse_mount_string_unix(string):
    """Handles parsing of mounts on Unix OS like Linux, Cygwin and MacOS."""
    parts = string.split(':')
    if len(parts) > 3:
        raise errors.InvalidArgument(
            f'Invalid mount format "{string}"'
        )
    if len(parts) == 1:
        return docker_types_services.Mount(target=parts[0], source=None)
    else:
        target = parts[1]
        source = parts[0]
        mount_type = 'volume'
        if source.startswith('/'):
            mount_type = 'bind'
        read_only = not (len(parts) == 2 or parts[2] == 'rw')
        return docker_types_services.Mount(target, source, read_only=read_only, type=mount_type)


def _parse_mount_string(string):
    """This is a fix to a bug in the Docker Python API by monkey patching the buggy method.

    The method is knowingly not supporting Windows. Until a fix PR is merged, we are hot patching the method to handle
    windows paths with : like C:/Users/bob/:/root.
    """
    if constants.IS_WINDOWS_PLATFORM:
        return _parse_mount_string_windows(string)
    else:
        return _parse_mount_string_unix(string)


docker_types_services.Mount.parse_mount_string = _parse_mount_string


class AgentRuntime:
    """Class to consolidate the agent settings and agent default definition, and create the agent service."""

    def __init__(self,
                 agent_settings: definitions.AgentSettings,
                 runtime_name: str,
                 docker_client: docker.DockerClient,
                 mq_service: mq.LocalRabbitMQ,
                 redis_service: redis.LocalRedis
                 ) -> None:
        """Constructs all the necessary attributes for the object.

        Args:
            agent: an agent definition containing all the settings of how agent should run and what arguments to pass.
            runtime_name: local runtime instance name.
            docker_client: docker client.
        """
        self._docker_client = docker_client
        self.agent = agent_settings
        self.image_name = agent_settings.container_image.split(':', maxsplit=1)[0]
        self.runtime_name = runtime_name
        self.mq_service = mq_service
        self.redis_service = redis_service
        self.update_agent_settings()

    def create_settings_config(self) -> docker.types.ConfigReference:
        """Create a docker configuration of the  agent settings.

        Returns:
            docker ConfigReference of the settings configuration
        """
        agent_instance_settings_proto = self.agent.to_raw_proto()
        config_name = hashlib.md5(f'config_settings_{self.image_name}_{self.runtime_name}'.encode()).hexdigest()

        try:
            settings_config = self._docker_client.configs.get(config_name)
            logging.warning('found existing config %s, config will removed', config_name)
            settings_config.remove()
        except docker.errors.NotFound:
            logging.debug('all good, config %s is new', config_name)

        docker_config = self._docker_client.configs.create(name=config_name,
                                                           labels={'ostorlab.universe': self.runtime_name},
                                                           data=agent_instance_settings_proto)
        return docker.types.ConfigReference(config_id=docker_config.id,
                                            config_name=config_name,
                                            filename='/tmp/settings.binproto')

    def create_definition_config(self) -> docker.types.ConfigReference:
        """Create a docker configuration of the  agent definition.

        Returns:
            docker configuration reference of the agent defintion configuration.
        """
        agent_definition = self._docker_client.images.get(self.agent.container_image).labels.get('agent_definition')
        config_name = hashlib.md5(f'config_definition_{self.image_name}__{self.runtime_name}'.encode()).hexdigest()

        try:
            settings_config = self._docker_client.configs.get(config_name)
            logging.warning('found existing config %s, config will removed', config_name)
            settings_config.remove()
        except docker.errors.NotFound:
            logging.debug('all good, config %s is new', config_name)

        docker_config = self._docker_client.configs.create(name=config_name,
                                                           labels={'ostorlab.universe': self.runtime_name},
                                                           data=str.encode(agent_definition))
        return docker.types.ConfigReference(config_id=docker_config.id,
                                            config_name=config_name,
                                            filename='/tmp/ostorlab.yaml')

    def create_agent_definition_from_label(self) -> agent_definitions.AgentDefinition:
        """Read the agent yaml definition from the docker image labels.

        Returns:
            the agent definition.
        """
        docker_image = self._docker_client.images.get(self.agent.container_image)
        yaml_definition_string = docker_image.labels.get('agent_definition')
        if yaml_definition_string is None:
            raise MissingAgentDefinitionLabel(f'agent definition label is missing from image {docker_image.tags[0]}')
        with io.StringIO(yaml_definition_string) as file:
            agent_definition = agent_definitions.AgentDefinition.from_yaml(file)
            return agent_definition

    def update_agent_settings(self) -> None:
        """Update agent settings with values from the local runtime."""
        self.agent.bus_url = self.mq_service.url
        self.agent.bus_exchange_topic = f'ostorlab_topic_{self.runtime_name}'
        self.agent.bus_management_url = self.mq_service.management_url
        self.agent.bus_vhost = self.mq_service.vhost
        self.agent.healthcheck_host = HEALTHCHECK_HOST
        self.agent.healthcheck_port = HEALTHCHECK_PORT
        self.agent.redis_url = self.redis_service.url

    def create_docker_healthchek(self) -> docker.types.Healthcheck:
        """Create a docker healthcheck configuration for the agent service.

        Returns:
            docker healthcheck configuration.
        """
        # wait 2s and check max 5 times with 0.5s between each check.
        healthcheck = docker.types.Healthcheck(test=['CMD', 'ostorlab', 'agent', 'healthcheck'],
                                               retries=HEALTHCHECK_RETRIES,
                                               timeout=HEALTHCHECK_TIMEOUT,
                                               start_period=HEALTHCHECK_START_PERIOD,
                                               interval=HEALTHCHECK_INTERVAL, )
        return healthcheck

    def replace_variable_mounts(self, mounts: List[str]):
        """Replace path variables for the container mounts

        Args:
            mounts: List of src:dst paths to mount
        """

        replaced_mounts = []
        for mount in mounts:
            for mount_variable, mount_value in MOUNT_VARIABLES.items():
                mount = mount.replace(mount_variable, mount_value)
            replaced_mounts.append(mount)
        return replaced_mounts

    def create_agent_service(self,
                             network_name: str,
                             extra_configs: Optional[List[docker.types.ConfigReference]] = None,
                             extra_mounts: Optional[List[docker.types.Mount]] = None
                             ) -> docker.models.services.Service:
        """Create the docker agent service with proper configs and policies.

        Args:
            network_name: network name to attach the service to.
            extra_configs: list of docker ConfigReferences that will be exposed to the service.

        Returns:
            the agent docker service.
        """
        agent_definition = self.create_agent_definition_from_label()
        self.agent.open_ports = self.agent.open_ports or agent_definition.open_ports
        if self.agent.open_ports:
            endpoint_spec = docker_types_services.EndpointSpec(
                ports={p.destination_port: p.source_port for p in self.agent.open_ports})
        else:
            endpoint_spec = docker_types_services.EndpointSpec(mode='dnsrr')

        configs = []
        configs.append(self.create_settings_config())
        configs.append(self.create_definition_config())
        if extra_configs is not None:
            configs.extend(extra_configs)

        mounts = self.agent.mounts or agent_definition.mounts
        mounts = self.replace_variable_mounts(mounts)
        if extra_mounts is not None:
            mounts.extend(extra_mounts)

        constraints = self.agent.constraints or agent_definition.constraints
        mem_limit = self.agent.mem_limit or agent_definition.mem_limit
        restart_policy = self.agent.restart_policy or agent_definition.restart_policy

        service_name = agent_definition.service_name or self.agent.container_image.replace(':', '_').\
            replace('.', '') + '_' + self.runtime_name
        agent_service = self._docker_client.services.create(
            image=self.agent.container_image,
            networks=[network_name],
            env=[
                f'UNIVERSE={self.runtime_name}',
            ],
            name=service_name,
            restart_policy=docker_types_services.RestartPolicy(condition=restart_policy),
            mounts=mounts,
            healthcheck=self.create_docker_healthchek(),
            labels={'ostorlab.universe': self.runtime_name},
            configs=configs,
            constraints=constraints,
            endpoint_spec=endpoint_spec,
            resources=docker_types_services.Resources(mem_limit=mem_limit))

        return agent_service



"""Asset of type agent for meta-scanning."""
import logging
from typing import Optional

import click

from ostorlab.assets import agent as agent_asset
from ostorlab.cli.scan.run import run

logger = logging.getLogger(__name__)


@run.run.command()
@click.option('--key', help='Agent key.', required=True)
@click.option('--version', help='Agent version.', required=False)
@click.option('--git-location', help='Agent source Git repo.', required=False)
@click.option('--docker-location', help='Agent store docker URL.', required=False)
@click.option('--yaml-file-location', help='Agent YAML definition path in the Git repo.', required=False)
@click.pass_context
def agent(ctx: click.core.Context, key: str, version: Optional[str] = None, git_location: Optional[str] = None,
          docker_location: Optional[str] = None, yaml_file_location: Optional[str] = None) -> None:
    """Run scan for agent."""
    runtime = ctx.obj['runtime']
    asset = agent_asset.Agent(key=key, version=version, git_location=git_location, docker_location=docker_location,
                              yaml_file_location=yaml_file_location)
    logger.debug('scanning asset %s', asset)
    runtime.scan(title=ctx.obj['title'], agent_group_definition=ctx.obj['agent_group_definition'], assets=[asset])



"""Module for the command list inside the group scan.
This module takes care of listing all the remote or local scans.
Example of usage:
    - ostorlab scan list --source=source."""

import click

from ostorlab.cli import console as cli_console
from ostorlab.cli.scan import scan
from ostorlab.utils import styles

console = cli_console.Console()


@scan.command(name='list')
@click.option('--page', '-p', help='Page number of scans you would like to see.', default=1)
@click.option('--elements', '-e', help='Number of scans to show per page.', default=10)
@click.pass_context
def list_scans(ctx: click.core.Context, page: int, elements: int) -> None:
    """List all your scans.\n
    Usage:\n
        - ostorlab scan --runtime=runtime list
    """
    runtime_instance = ctx.obj['runtime']
    with console.status('Fetching scans'):
        scans = runtime_instance.list(page=page, number_elements=elements)
        if scans is not None:
            console.success('Scans listed successfully.')
            columns = {'Id': 'id',
                       'Asset': 'asset',
                       'Created Time': 'created_time',
                       'Progress': 'progress'}
            title = f'Showing {len(scans)} Scans'

            data = [
                {
                    'id': str(s.id),
                    'asset': s.asset,
                    'created_time': str(s.created_time),
                    'progress': styles.style_progress(s.progress),
                } for s in scans]

            console.table(columns=columns, data=data, title=title)
        else:
            console.error('Error fetching scan list.')



"""Get agent details from the public api."""

import json
from typing import Dict, Optional

from ostorlab.apis import request

class AgentDetailsAPIRequest(request.APIRequest):
    """Get agent details for a specified agent_key."""

    def __init__(self, agent_key: str) -> None:
        """Initializer"""
        self._agent_key = agent_key

    @property
    def query(self) -> Optional[str]:
        """The query to fetch the agent details with an agent key.

        Returns:
            The query to fetch the agent details.
        """
        return """
            query Agent($agentKey: String!){
                agent(agentKey: $agentKey) {
                    name,
                    gitLocation,
                    yamlFileLocation,
                    dockerLocation,
                    access,
                    listable,
                    key
                    versions(orderBy: Version, sort: Desc, page: 1, numberElements: 1) {
                      versions {
                        version
                      }
                    }
                }
            }
        """

    @property
    def data(self) -> Optional[Dict]:
        """Sets the body of the API request, to fetch the specific agent.

        Returns:
            The body of the agent details request.
        """
        data = {
            'query': self.query,
            'variables': json.dumps({'agentKey': self._agent_key})
        }
        return data



"""Agent List command."""
import logging

import docker

from ostorlab.cli import console as cli_console
from ostorlab.cli.agent import agent

console = cli_console.Console()

logger = logging.getLogger(__name__)


def _human_readable_size(size, decimal_places=3):
    for unit in ['B', 'KiB', 'MiB', 'GiB', 'TiB']:
        if size < 1024.0:
            break
        size /= 1024.0
    return f'{size:.{decimal_places}f}{unit}'


@agent.command(name='list')
def list_cli() -> None:
    """CLI command to list installed agents."""
    docker_client = docker.from_env()
    images = docker_client.images.list()
    console.success('Agents listed successfully.')
    agents = []
    for im in images:
        for t in im.tags:
            if t.startswith('agent_'):
                agents.append({
                    'id': im.short_id,
                    'agent': t.replace('_', '/', 2).split(':')[0],
                    'version': t.split(':')[1],
                    'size': _human_readable_size(im.attrs['Size']),
                    'created': im.attrs['Created']
                })

    columns = {
        'Agent': 'agent',
        'Version': 'version',
        'Id': 'id',
        'Size': 'size',
        'Created': 'created',
    }
    title = f'Found {len(agents)} Agents'
    console.table(columns=columns, data=agents, title=title)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: agent_instance_settings.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='agent_instance_settings.proto',
  package='ostorlab.runtimes.proto',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n\x1d\x61gent_instance_settings.proto\x12\x17ostorlab.runtimes.proto\"0\n\x03\x41rg\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\x0c\"<\n\x0bPortMapping\x12\x13\n\x0bsource_port\x18\x01 \x01(\r\x12\x18\n\x10\x64\x65stination_port\x18\x02 \x01(\r\"\x8f\x03\n\x15\x41gentInstanceSettings\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x0f\n\x07\x62us_url\x18\x02 \x01(\t\x12\x1a\n\x12\x62us_exchange_topic\x18\x03 \x01(\t\x12\x1a\n\x12\x62us_management_url\x18\x04 \x01(\t\x12\x11\n\tbus_vhost\x18\x05 \x01(\t\x12*\n\x04\x61rgs\x18\x06 \x03(\x0b\x32\x1c.ostorlab.runtimes.proto.Arg\x12\x13\n\x0b\x63onstraints\x18\x07 \x03(\t\x12\x0e\n\x06mounts\x18\x08 \x03(\t\x12\x16\n\x0erestart_policy\x18\t \x01(\t\x12\x11\n\tmem_limit\x18\n \x01(\x04\x12\x38\n\nopen_ports\x18\x0b \x03(\x0b\x32$.ostorlab.runtimes.proto.PortMapping\x12\x10\n\x08replicas\x18\x0c \x01(\r\x12\x18\n\x10healthcheck_host\x18\r \x01(\t\x12\x18\n\x10healthcheck_port\x18\x0e \x01(\r\x12\x11\n\tredis_url\x18\x0f \x01(\t')
)




_ARG = _descriptor.Descriptor(
  name='Arg',
  full_name='ostorlab.runtimes.proto.Arg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='ostorlab.runtimes.proto.Arg.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='ostorlab.runtimes.proto.Arg.type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='ostorlab.runtimes.proto.Arg.value', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=58,
  serialized_end=106,
)


_PORTMAPPING = _descriptor.Descriptor(
  name='PortMapping',
  full_name='ostorlab.runtimes.proto.PortMapping',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='source_port', full_name='ostorlab.runtimes.proto.PortMapping.source_port', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='destination_port', full_name='ostorlab.runtimes.proto.PortMapping.destination_port', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=108,
  serialized_end=168,
)


_AGENTINSTANCESETTINGS = _descriptor.Descriptor(
  name='AgentInstanceSettings',
  full_name='ostorlab.runtimes.proto.AgentInstanceSettings',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='ostorlab.runtimes.proto.AgentInstanceSettings.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bus_url', full_name='ostorlab.runtimes.proto.AgentInstanceSettings.bus_url', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bus_exchange_topic', full_name='ostorlab.runtimes.proto.AgentInstanceSettings.bus_exchange_topic', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bus_management_url', full_name='ostorlab.runtimes.proto.AgentInstanceSettings.bus_management_url', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bus_vhost', full_name='ostorlab.runtimes.proto.AgentInstanceSettings.bus_vhost', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='args', full_name='ostorlab.runtimes.proto.AgentInstanceSettings.args', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='constraints', full_name='ostorlab.runtimes.proto.AgentInstanceSettings.constraints', index=6,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mounts', full_name='ostorlab.runtimes.proto.AgentInstanceSettings.mounts', index=7,
      number=8, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='restart_policy', full_name='ostorlab.runtimes.proto.AgentInstanceSettings.restart_policy', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mem_limit', full_name='ostorlab.runtimes.proto.AgentInstanceSettings.mem_limit', index=9,
      number=10, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='open_ports', full_name='ostorlab.runtimes.proto.AgentInstanceSettings.open_ports', index=10,
      number=11, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='replicas', full_name='ostorlab.runtimes.proto.AgentInstanceSettings.replicas', index=11,
      number=12, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='healthcheck_host', full_name='ostorlab.runtimes.proto.AgentInstanceSettings.healthcheck_host', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='healthcheck_port', full_name='ostorlab.runtimes.proto.AgentInstanceSettings.healthcheck_port', index=13,
      number=14, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='redis_url', full_name='ostorlab.runtimes.proto.AgentInstanceSettings.redis_url', index=14,
      number=15, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=171,
  serialized_end=570,
)

_AGENTINSTANCESETTINGS.fields_by_name['args'].message_type = _ARG
_AGENTINSTANCESETTINGS.fields_by_name['open_ports'].message_type = _PORTMAPPING
DESCRIPTOR.message_types_by_name['Arg'] = _ARG
DESCRIPTOR.message_types_by_name['PortMapping'] = _PORTMAPPING
DESCRIPTOR.message_types_by_name['AgentInstanceSettings'] = _AGENTINSTANCESETTINGS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Arg = _reflection.GeneratedProtocolMessageType('Arg', (_message.Message,), dict(
  DESCRIPTOR = _ARG,
  __module__ = 'agent_instance_settings_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.runtimes.proto.Arg)
  ))
_sym_db.RegisterMessage(Arg)

PortMapping = _reflection.GeneratedProtocolMessageType('PortMapping', (_message.Message,), dict(
  DESCRIPTOR = _PORTMAPPING,
  __module__ = 'agent_instance_settings_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.runtimes.proto.PortMapping)
  ))
_sym_db.RegisterMessage(PortMapping)

AgentInstanceSettings = _reflection.GeneratedProtocolMessageType('AgentInstanceSettings', (_message.Message,), dict(
  DESCRIPTOR = _AGENTINSTANCESETTINGS,
  __module__ = 'agent_instance_settings_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.runtimes.proto.AgentInstanceSettings)
  ))
_sym_db.RegisterMessage(AgentInstanceSettings)


# @@protoc_insertion_point(module_scope)



"""Link asset."""
import dataclasses

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector('v3.asset.link')
class Link(asset.Asset):
    """Agent asset."""
    url: str
    method: str

    def __str__(self):
        return f'Link {self.url} with method {self.method}'



"""Ostorlab CLI agent install command."""

import click

from ostorlab.cli.agent import agent as agent_command
from ostorlab.cli import console as cli_console
from ostorlab.cli import install_agent
from ostorlab.cli import docker_requirements_checker


console = cli_console.Console()


@agent_command.command()
@click.argument('agent', required=True)
@click.option('--version', '-v', help='Agent version.', required=False)
def install(agent: str, version: str = '') -> None:
    """Install an agent : pull the image from the ostorlab store."""

    if not docker_requirements_checker.is_docker_installed():
        console.error('Docker is not installed.')
        raise click.exceptions.Exit(2)
    elif not docker_requirements_checker.is_sys_arch_supported():
        console.error('System architecture is not supported.')
        raise click.exceptions.Exit(2)
    elif not docker_requirements_checker.is_user_permitted():
        console.error('User does not have permissions to run docker.')
        raise click.exceptions.Exit(2)
    elif not docker_requirements_checker.is_docker_working():
        console.error('Error using docker.')
        raise click.exceptions.Exit(2)
    else:
        try:
            install_agent.install(agent, version)
        except install_agent.AgentDetailsNotFound as e:
            console.error(e)
            raise click.exceptions.Exit(2)



"""Agent health check mixin to ensure agents are operational.

The mixin runs a webservice by default on 0.0.0.0:5000 that must return 200 status and OK as a response.
The mixin enables adding multiple health check callbacks that must all return True.

Typical usage:

```python
    status_agent = agent_healthcheck_mixin.AgentHealthcheckMixin()
    status_agent.add_healthcheck(self._is_healthy)
    status_agent.start()
```
"""
import logging
from threading import Thread
from typing import Optional, Callable

import flask
from werkzeug import serving

logger = logging.getLogger(__name__)

DEFAULT_PORT = 5000
DEFAULT_HOST = '0.0.0.0'


class HealthcheckWebThread(Thread):
    """Webservice thread that return OK and NOK dependencies the health check return."""

    def __init__(self, name: str, host: str, port: int) -> None:
        """Setups a flask server.

        Args:
            name: name of the Flask service.
            host: host on which the web service is listening.
            port: port on which the web service is listening.
        """
        super().__init__()
        logger.info('Preparing flask')
        self._app = flask.Flask(name)
        self._server = serving.make_server(host=host, port=port, app=self._app, threaded=True)
        self._healthcheck_callbacks = []
        self._disable_verbose_logging()

    def _disable_verbose_logging(self):
        """Disable Flaskserver verbose logging."""
        log = logging.getLogger('werkzeug')
        log.setLevel(logging.ERROR)

    def run(self) -> None:
        """Starts a long-running web service.

        Returns:
            None
        """
        logger.info('starting status server')
        self._server.serve_forever()

    def start(self) -> None:
        """Start thread in deamon mode.

        Returns:
            None
        """
        self._add_urls()
        self.daemon = True
        super().start()

    def stop(self) -> None:
        """Stop thread.

        Returns:
            None
        """
        self._server.shutdown()

    def _add_urls(self) -> None:
        """Add status URL at /stats."""
        self._app.add_url_rule('/status', 'status', self._status)

    def add_healthcheck(self, healthcheck_callback: Callable[[], bool]) -> None:
        """Add health check call back function that status will evaluate when called.

        Args:
            healthcheck_callback: Callback function, must take no argument and return a bool.

        Returns:
            None
        """
        logger.info('Adding healthcheck callback')
        self._healthcheck_callbacks.append(healthcheck_callback)

    def _status(self):
        """Healthcheck status endpoint.

        Returns:
            OK if all checks are work, NOK if not.
        """
        if all(healthcheck() for healthcheck in self._healthcheck_callbacks):
            logger.debug('Health checks status OK')
            return 'OK'
        else:
            logger.error('Health checks status NOK')
            return 'NOK'


class AgentHealthcheckMixin:
    """Agent heath check mixin exposing a pull-style health check function as a web endpoint.

    The endpoint is listing on http://0.0.0.0:5000/status and will return OK if all is good, and NOK in case one of the
    checks fail.

    **IMPORTANT**: the check might fail in a different way, for instance in case a check callback throws an exception,
    the web service intentionally do not catch them to make it easier to detect and debug. Health check should check
    for the presence of the OK and not the absence of NOK.
    """

    def __init__(self, name: Optional[str] = None, host=DEFAULT_HOST, port=DEFAULT_PORT) -> None:
        """Inits the health check web service thread and provides sane defaults.

        Args:
            name: name of the Flask service.
            host: host on which the web service is listening, defaults to 0.0.0.0.
            port: port on which the web service is listening, defaults to 5000.
        """
        self._healthcheck_web_thread = HealthcheckWebThread(name=(name or __name__), host=host, port=port)
        logger.debug('Starting healthcheck for agent.')

    def add_healthcheck(self, healthcheck_callback: Callable[[], bool]) -> None:
        """Add health check call back function that status will evaluate when called.

        Args:
            healthcheck_callback: Callback function, must take no argument and return a bool.

        Returns:
            None
        """
        logger.debug('enabling healthcheck')
        self._healthcheck_web_thread.add_healthcheck(healthcheck_callback)

    def start_healthcheck(self) -> None:
        """Start exposing the health service.

        Returns:
            None
        """
        self._healthcheck_web_thread.start()
        logger.debug('Healthcheck web thread started.')

    def stop_healthcheck(self) -> None:
        """Stop exposing the health service.

        Returns:
            None
        """
        self._healthcheck_web_thread.stop()
        logger.debug('Healthcheck web thread stopped.')



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: v3/report/vulnerability/vulnerability.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='v3/report/vulnerability/vulnerability.proto',
  package='v3.report.event.vulnerability',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n+v3/report/vulnerability/vulnerability.proto\x12\x1dv3.report.event.vulnerability\"\'\n\tReference\x12\r\n\x05title\x18\x01 \x01(\t\x12\x0b\n\x03url\x18\x02 \x01(\t\"\xf1\x03\n\x07Message\x12\r\n\x05title\x18\x01 \x01(\t\x12>\n\x0brisk_rating\x18\x02 \x01(\x0e\x32).v3.report.event.vulnerability.RiskRating\x12\x18\n\x10technical_detail\x18\x03 \x01(\t\x12\x19\n\x11short_description\x18\x04 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12\x16\n\x0erecommendation\x18\x06 \x01(\t\x12<\n\nreferences\x18\x07 \x03(\x0b\x32(.v3.report.event.vulnerability.Reference\x12\x1d\n\x0esecurity_issue\x18\x08 \x01(\x08:\x05\x66\x61lse\x12\x1c\n\rprivacy_issue\x18\t \x01(\x08:\x05\x66\x61lse\x12!\n\x12has_public_exploit\x18\n \x01(\x08:\x05\x66\x61lse\x12\"\n\x13targeted_by_malware\x18\x0b \x01(\x08:\x05\x66\x61lse\x12%\n\x16targeted_by_ransomware\x18\x0c \x01(\x08:\x05\x66\x61lse\x12\'\n\x18targeted_by_nation_state\x18\r \x01(\x08:\x05\x66\x61lse\x12\x0b\n\x03\x64na\x18\x0e \x01(\t\x12\x16\n\x0e\x63vss_v3_vector\x18\x0f \x01(\t*p\n\nRiskRating\x12\x08\n\x04HIGH\x10\x00\x12\n\n\x06MEDIUM\x10\x01\x12\x07\n\x03LOW\x10\x02\x12\x0f\n\x0bPOTENTIALLY\x10\x03\x12\r\n\tHARDENING\x10\x04\x12\n\n\x06SECURE\x10\x05\x12\r\n\tIMPORTANT\x10\x06\x12\x08\n\x04INFO\x10\x07')
)

_RISKRATING = _descriptor.EnumDescriptor(
  name='RiskRating',
  full_name='v3.report.event.vulnerability.RiskRating',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='HIGH', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MEDIUM', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOW', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='POTENTIALLY', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HARDENING', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SECURE', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IMPORTANT', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INFO', index=7, number=7,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=619,
  serialized_end=731,
)
_sym_db.RegisterEnumDescriptor(_RISKRATING)

RiskRating = enum_type_wrapper.EnumTypeWrapper(_RISKRATING)
HIGH = 0
MEDIUM = 1
LOW = 2
POTENTIALLY = 3
HARDENING = 4
SECURE = 5
IMPORTANT = 6
INFO = 7



_REFERENCE = _descriptor.Descriptor(
  name='Reference',
  full_name='v3.report.event.vulnerability.Reference',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='title', full_name='v3.report.event.vulnerability.Reference.title', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='url', full_name='v3.report.event.vulnerability.Reference.url', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=78,
  serialized_end=117,
)


_MESSAGE = _descriptor.Descriptor(
  name='Message',
  full_name='v3.report.event.vulnerability.Message',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='title', full_name='v3.report.event.vulnerability.Message.title', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='risk_rating', full_name='v3.report.event.vulnerability.Message.risk_rating', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='technical_detail', full_name='v3.report.event.vulnerability.Message.technical_detail', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='short_description', full_name='v3.report.event.vulnerability.Message.short_description', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='v3.report.event.vulnerability.Message.description', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='recommendation', full_name='v3.report.event.vulnerability.Message.recommendation', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='references', full_name='v3.report.event.vulnerability.Message.references', index=6,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='security_issue', full_name='v3.report.event.vulnerability.Message.security_issue', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='privacy_issue', full_name='v3.report.event.vulnerability.Message.privacy_issue', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='has_public_exploit', full_name='v3.report.event.vulnerability.Message.has_public_exploit', index=9,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='targeted_by_malware', full_name='v3.report.event.vulnerability.Message.targeted_by_malware', index=10,
      number=11, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='targeted_by_ransomware', full_name='v3.report.event.vulnerability.Message.targeted_by_ransomware', index=11,
      number=12, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='targeted_by_nation_state', full_name='v3.report.event.vulnerability.Message.targeted_by_nation_state', index=12,
      number=13, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dna', full_name='v3.report.event.vulnerability.Message.dna', index=13,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cvss_v3_vector', full_name='v3.report.event.vulnerability.Message.cvss_v3_vector', index=14,
      number=15, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=120,
  serialized_end=617,
)

_MESSAGE.fields_by_name['risk_rating'].enum_type = _RISKRATING
_MESSAGE.fields_by_name['references'].message_type = _REFERENCE
DESCRIPTOR.message_types_by_name['Reference'] = _REFERENCE
DESCRIPTOR.message_types_by_name['Message'] = _MESSAGE
DESCRIPTOR.enum_types_by_name['RiskRating'] = _RISKRATING
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Reference = _reflection.GeneratedProtocolMessageType('Reference', (_message.Message,), dict(
  DESCRIPTOR = _REFERENCE,
  __module__ = 'v3.report.vulnerability.vulnerability_pb2'
  # @@protoc_insertion_point(class_scope:v3.report.event.vulnerability.Reference)
  ))
_sym_db.RegisterMessage(Reference)

Message = _reflection.GeneratedProtocolMessageType('Message', (_message.Message,), dict(
  DESCRIPTOR = _MESSAGE,
  __module__ = 'v3.report.vulnerability.vulnerability_pb2'
  # @@protoc_insertion_point(class_scope:v3.report.event.vulnerability.Message)
  ))
_sym_db.RegisterMessage(Message)


# @@protoc_insertion_point(module_scope)



"""Agent Healthcheck commands."""
import requests
import logging

import click

from ostorlab.cli import console as cli_console
from ostorlab.cli.agent import agent

console = cli_console.Console()

logger = logging.getLogger(__name__)


@agent.command()
@click.option('--host', '-h', help='Host to check healthy agent.', required=False, default='localhost')
@click.option('--port', '-p', help='Port to check healthy agent.', required=False, default=5000, type=int)
@click.option('--https/--no-https', help='Enable HTTPS.', required=False, default=False)
def healthcheck(host: str, port: int, https: bool = False) -> None:
    """Minimal agent healthcheck command to ensure the agent response with 200 OK on localhost:5000 by default."""
    if https:
        url = f'https://{host}:{port}/status'
    else:
        url = f'http://{host}:{port}/status'

    try:
        response = requests.get(url)
        if response.status_code != 200:
            console.error(f'Response status code is {response.status_code}')
            raise click.exceptions.Exit(2)
        else:
            console.success(f'Healthcheck OK on {url}!')
    except requests.ConnectionError as e:
        console.error(f'Error checking agent health on {url}')
        raise click.exceptions.Exit(2) from e



"""Redis service in charge of providing capabilities like QPS limiting, distributed locking and distributed storage."""
import logging
from typing import Dict

import docker
import tenacity
from docker import errors
from docker import types
from docker.models import services

logger = logging.getLogger(__name__)

REDIS_IMAGE = 'redis:latest'


class LocalRedis:
    """Redis service spawned a docker swarm service."""

    def __init__(self,
                 name: str,
                 network: str,
                 exposed_ports: Dict[int, int] = None,
                 image: str = REDIS_IMAGE) -> None:
        """Initialize the Redis service parameters.
        Args:
            name: Name of the service.
            network: Network used for the Docker Redis service.
            exposed_ports: The list of Redis service exposed ports
            image: Redis Docker image
        """
        self._name = name
        self._docker_client = docker.from_env()
        # images
        self._redis_image = image
        self._network = network
        self._redis_host = f'redis_{self._name}'
        # service
        self._redis_service = None
        # exposed_port
        self._exposed_ports = exposed_ports

    @property
    def url(self) -> str:
        """URL to connect to the local Redis instance."""
        return f'redis://{self._redis_host}:6379/'

    @property
    def service(self):
        return self._redis_service

    def start(self) -> None:
        """Start local Redis instance."""
        self._create_network()
        self._redis_service = self._start_redis()

        if self._redis_service is not None and not self._is_service_healthy():
            logger.error('Redis container for service %s is not ready', self._redis_service.id)
            return

    def stop(self):
        """Stop the local Redis instance."""
        for service in self._docker_client.services.list():
            universe = service.attrs['Spec']['Labels'].get('ostorlab.universe')
            if universe is not None and service.name.startswith('redis_') and self._name in universe:
                service.remove()

    def _create_network(self):
        if any(network.name == self._network for network in self._docker_client.networks.list()):
            logger.warning('network already exists.')
        else:
            logger.info('creating private network %s', self._network)
            return self._docker_client.networks.create(
                name=self._network,
                driver='overlay',
                attachable=True,
                labels={'ostorlab.universe': self._name},
                check_duplicate=True
            )

    def _start_redis(self) -> services.Service:
        try:
            logger.info('starting Redis')
            endpoint_spec = types.services.EndpointSpec(mode='vip', ports=self._exposed_ports)
            service_mode = types.services.ServiceMode('replicated', replicas=1)
            return self._docker_client.services.create(
                image=self._redis_image,
                networks=[self._network],
                name=self._redis_host,
                restart_policy=types.RestartPolicy(condition='any'),
                mode=service_mode,
                labels={'ostorlab.universe': self._name, 'ostorlab.redis': ''},
                endpoint_spec=endpoint_spec)
        except docker.errors.APIError as e:
            error_message = f'Redis service could not be started. Reason: {e}.'
            logger.error(error_message)
            return

    @tenacity.retry(stop=tenacity.stop_after_attempt(20),
                    wait=tenacity.wait_exponential(multiplier=1, max=12),
                    # return last value and don't raise RetryError exception.
                    retry_error_callback=lambda lv: lv.result(),
                    retry=tenacity.retry_if_result(lambda v: v is False))
    def _is_service_healthy(self) -> bool:
        logger.info('checking service %s', self._redis_service.name)
        return self.is_healthy

    @property
    def is_healthy(self) -> bool:
        try:
            return self._redis_service is not None and \
                   len([task for task in self._redis_service.tasks() if task['Status']['State'] == 'running']) == 1
        except errors.DockerException:
            return False
